/*
    Scsign 
    Copyleft (C) 2004 Antonino Iacono

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
** include functions from
** Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
**
** include functions from 
** OpenSC (www.opensc.org).
**
** Thanks And Credits to Diska (discardin@tiscali.it)
*/

#include "scsign.h"

extern char* certificato;
extern char *file_certificato;
extern char *opt_pincode;

static int init(EVP_MD_CTX *ctx)
        { return SHA1_Init(ctx->md_data); }

static int update(EVP_MD_CTX *ctx,const void *data,unsigned long count)
        { return SHA1_Update(ctx->md_data,data,count); }

static int final(EVP_MD_CTX *ctx,unsigned char *md)
        { return SHA1_Final(md,ctx->md_data); }


static EVP_MD sha1a_md=
        {
        NID_sha1,
        NID_sha1WithRSAEncryption,
        SHA_DIGEST_LENGTH,
        0,
        init,
        update,
        final,
        NULL,
        NULL,
        EVP_PKEY_RSA_method2,
        SHA_CBLOCK,
        sizeof(EVP_MD *)+sizeof(SHA_CTX),
        };

EVP_MD *EVP_sha1a(void)
        {
        return(&sha1a_md);
        }

struct sc_context *ctx = NULL;
struct sc_card *card = NULL;

char opt_input[9];
struct sc_pkcs15_card *p15card = NULL;
int opt_crypt_flags = 0;
unsigned char hash_sign[129];
unsigned char hash[20];
void die();

int mysign(char *infile, char* signerfile)
	{
	X509 *x509;
	EVP_PKEY *pkey;
	RSA *rsa;
	PKCS7 *p7;
	PKCS7_SIGNER_INFO *si;
	BIO *data1,*p7bio;
	char buf[1024*4];
	int i;
	BIO *in = NULL, *out = NULL;
        char * outfile[64];

        strcpy((char *)outfile,infile);
        strcat((char *)outfile,".p7m"); 


#ifndef NO_SHA1
	EVP_add_digest(EVP_sha1a());
#endif
	
EVP_add_cipher(EVP_rc2_cbc());

	data1=BIO_new(BIO_s_file());

                if ((pkey=EVP_PKEY_new()) == NULL)
                        {
                        abort();
                        goto err;
                        }

		rsa=RSA_generate_key(1024,RSA_F4,NULL,NULL);
        	
		if (!EVP_PKEY_assign_RSA(pkey,rsa))
                {
                abort();
                goto err;
                }


if (!BIO_read_filename(data1,infile))	goto err;
if (!(out= BIO_new_file((char *)outfile,"w"))) goto err;

if ((in=BIO_new_file(signerfile,"rb")) == NULL) goto err;
if ((x509=PEM_read_bio_X509(in,NULL,NULL,NULL)) == NULL) goto err;
BIO_reset(in);
BIO_free(in);

	p7=PKCS7_new();
	PKCS7_set_type(p7,NID_pkcs7_signed);
	 
	si=PKCS7_add_signature(p7,x509,pkey,EVP_sha1a());
	if (si == NULL) goto err;

	/* If you do this then you get signing time automatically added */
	PKCS7_add_signed_attribute(si, NID_pkcs9_contentType, V_ASN1_OBJECT,
						OBJ_nid2obj(NID_pkcs7_data));

	PKCS7_add_certificate(p7,x509);

	PKCS7_content_new(p7,NID_pkcs7_data);

	if ((p7bio=PKCS7_dataInit(p7,NULL)) == NULL) goto err;

	for (;;)
		{
		i=BIO_read(data1,buf,sizeof(buf));
		if (i <= 0) break;
		BIO_write(p7bio,buf,i);
		}

	if (!PKCS7_dataFinal(p7,p7bio)) goto err;
	BIO_free(p7bio);

	i2d_PKCS7_bio(out,p7);

	PKCS7_free(p7);
	BIO_free_all(out);
return 0;

err:
	ERR_load_crypto_strings();
return 1;
	}

int RSA_eay_private_encrypt2(int flen, const unsigned char *from,
	     unsigned char *to, RSA *rsa, int padding)
	{

	unsigned char *buf=NULL;
	BIGNUM f,ret;
	int i,j,num=0,in,r= -1;
	BN_CTX *ctx=NULL;
	BN_init(&f);
	BN_init(&ret);

	if ((ctx=BN_CTX_new()) == NULL) goto err;
	num=BN_num_bytes(rsa->n);
	if ((buf=(unsigned char *)OPENSSL_malloc(num)) == NULL)
		{
		RSAerr(RSA_F_RSA_EAY_PRIVATE_ENCRYPT,ERR_R_MALLOC_FAILURE);
		goto err;
		}

	switch (padding)
		{
	case RSA_PKCS1_PADDING:
		i=RSA_padding_add_PKCS1_type_1(buf,num,from,flen);
		break;
	case RSA_NO_PADDING:
		i=RSA_padding_add_none(buf,num,from,flen);
		break;
	case RSA_SSLV23_PADDING:
	default:
		RSAerr(RSA_F_RSA_EAY_PRIVATE_ENCRYPT,RSA_R_UNKNOWN_PADDING_TYPE);
		goto err;
		}

     
	for (in=108;in<128;in++) hash[in-108]=buf[in]; 

	if (i <= 0) goto err;

	if (BN_bin2bn(buf,num,&f) == NULL) goto err;
	
	if (BN_ucmp(&f, rsa->n) >= 0) goto err;

	if ((rsa->flags & RSA_FLAG_BLINDING) && (rsa->blinding == NULL))
		RSA_blinding_on(rsa,ctx);
	if (rsa->flags & RSA_FLAG_BLINDING)
		if (!BN_BLINDING_convert(&f,rsa->blinding,ctx)) goto err;

	if (rsa->flags & RSA_FLAG_BLINDING)
		if (!BN_BLINDING_invert(&ret,rsa->blinding,ctx)) goto err;

	j=BN_num_bytes(&ret);
	i=BN_bn2bin(&ret,&(to[num-j]));

        if(!p15sign(hash))  memmove(to,hash_sign, 128 * sizeof( unsigned char ) ); 

	r=num;

err:
	if (ctx != NULL) BN_CTX_free(ctx);
	BN_clear_free(&ret);
	BN_clear_free(&f);
	if (buf != NULL)
		{
		memset(buf,0,num);
		OPENSSL_free(buf);
		}

	return(r);
	}


int RSA_sign2(int type, const unsigned char *m, unsigned int m_len,unsigned char *sigret, unsigned int *siglen, RSA *rsa)
	{
	X509_SIG sig;
	ASN1_TYPE parameter;
	int i,j,ret=1;
	unsigned char *p, *tmps = NULL;
	const unsigned char *s = NULL;
	X509_ALGOR algor;
	ASN1_OCTET_STRING digest;
	if((rsa->flags & RSA_FLAG_SIGN_VER) && rsa->meth->rsa_sign)
		{
		return rsa->meth->rsa_sign(type, m, m_len,
			sigret, siglen, rsa);
		}

	if(type == NID_md5_sha1) {
		if(m_len != SSL_SIG_LENGTH) {
			RSAerr(RSA_F_RSA_SIGN,RSA_R_INVALID_MESSAGE_LENGTH);
			return(0);
		}
		i = SSL_SIG_LENGTH;
		s = m;
	} else {
		sig.algor= &algor;
		sig.algor->algorithm=OBJ_nid2obj(type);
		if (sig.algor->algorithm == NULL)
			{
			RSAerr(RSA_F_RSA_SIGN,RSA_R_UNKNOWN_ALGORITHM_TYPE);
			return(0);
			}
		if (sig.algor->algorithm->length == 0)
			{
			RSAerr(RSA_F_RSA_SIGN,RSA_R_THE_ASN1_OBJECT_IDENTIFIER_IS_NOT_KNOWN_FOR_THIS_MD);
			return(0);
			}
		parameter.type=V_ASN1_NULL;
		parameter.value.ptr=NULL;
		sig.algor->parameter= &parameter;

		sig.digest= &digest;
		sig.digest->data=(unsigned char *)m; /* TMP UGLY CAST */
		sig.digest->length=m_len;

		i=i2d_X509_SIG(&sig,NULL);
	}
	j=RSA_size(rsa);
	if (i > (j-RSA_PKCS1_PADDING_SIZE))
		{
		RSAerr(RSA_F_RSA_SIGN,RSA_R_DIGEST_TOO_BIG_FOR_RSA_KEY);
		return(0);
		}
	if(type != NID_md5_sha1) {
		tmps=(unsigned char *)OPENSSL_malloc((unsigned int)j+1);
		if (tmps == NULL)
			{
			RSAerr(RSA_F_RSA_SIGN,ERR_R_MALLOC_FAILURE);
			return(0);
			}
		p=tmps;
		i2d_X509_SIG(&sig,&p);
		s=tmps;
	}
	i=RSA_eay_private_encrypt2(i,s,sigret,rsa,RSA_PKCS1_PADDING);
	if (i <= 0)
		ret=0;
	else
		*siglen=i;

	if(type != NID_md5_sha1) {
		OPENSSL_cleanse(tmps,(unsigned int)j+1);
		OPENSSL_free(tmps);
	}
	return(ret);
	}

int p15sign(unsigned char* hash)
{       
	struct sc_context *ctx = NULL;
        struct sc_card *card = NULL;
	struct sc_pkcs15_object *key;
	int r;	

	opt_crypt_flags |= SC_ALGORITHM_RSA_HASH_SHA1;
	opt_crypt_flags |= SC_ALGORITHM_RSA_PAD_PKCS1;
    
	r = connect_pcsc();
	if (r) goto end;

		if ((get_key(SC_PKCS15_PRKEY_USAGE_SIGN|
			   SC_PKCS15_PRKEY_USAGE_SIGNRECOVER|
			   SC_PKCS15_PRKEY_USAGE_NONREPUDIATION, &key))
		 || (sign(key,hash))) fprintf(stderr, "Error signing\n");

end:
 die();

return r;

}                       

int connect_card(sc_context_t *ctx, sc_card_t **cardp,
		 int reader_id, int slot_id, int wait, int quiet)
{
	sc_reader_t *reader;
	sc_card_t *card;
	int r;

	if (wait) {
		struct sc_reader *readers[16];
		int slots[16];
		int i, j, k, found;
		unsigned int event;

		for (i = k = 0; i < ctx->reader_count; i++) {
			if (reader_id >= 0 && reader_id != i)
				continue;
			reader = ctx->reader[i];
			for (j = 0; j < reader->slot_count; j++, k++) {
				readers[k] = reader;
				slots[k] = j;
			}
		}

		r = sc_wait_for_event(readers, slots, k,
				SC_EVENT_CARD_INSERTED,
				&found, &event, -1);
		if (r < 0) {
			fprintf(stderr,
				"Error while waiting for card: %s\n",
				sc_strerror(r));
			return r;
		}

		reader = readers[found];
		slot_id = slots[found];
	} else {
		if (reader_id < 0)
			reader_id = 0;
		if (ctx->reader_count == 0) {
			fprintf(stderr,
				"No smart card readers configured.\n");
			return 1;
		}
		if (reader_id >= ctx->reader_count) {
			fprintf(stderr,
				"Illegal reader number. "
				"Only %d reader(s) configured.\n",
				ctx->reader_count);
			return 1;
		}

		reader = ctx->reader[reader_id];
		slot_id = 0;
		if (sc_detect_card_presence(reader, 0) <= 0) {
			fprintf(stderr, "Card not present.\n");
			return 3;
		}
	}

	if (!quiet)
		printf("Connecting to card in reader %s...\n", reader->name);
	if ((r = sc_connect_card(reader, slot_id, &card)) < 0) {
		fprintf(stderr,
			"Failed to connect to card: %s\n",
			sc_strerror(r));
		return r;
	}

	if (!quiet)
		printf("Using card driver %s.\n", card->driver->name);

	if ((r = sc_lock(card)) < 0) {
		fprintf(stderr,
			"Failed to lock card: %s\n",
			sc_strerror(r));
		sc_disconnect_card(card, 3);
		return r;
	}

	*cardp = card;
	return 0;
}

int sign(struct sc_pkcs15_object *obj, unsigned char *hash)
{
	u8 out[256];
	struct sc_pkcs15_prkey_info *key = (struct sc_pkcs15_prkey_info *) obj->data;
	int r, len, i;
	
	len = sizeof(out);
	if (obj->type == SC_PKCS15_TYPE_PRKEY_RSA
	 && !(opt_crypt_flags & SC_ALGORITHM_RSA_PAD_PKCS1)) {
		
		fprintf(stderr, "Input has to be exactly %d bytes, when using no padding.\n",
			key->modulus_length/8);
		return 2;
	}
	if (!key->native) {

		fprintf(stderr, "Cannot use extractable key because this "
				"program was compiled without crypto "
				"support.\n");

		r = SC_ERROR_NOT_SUPPORTED;
	} else {
		r = sc_pkcs15_compute_signature(p15card, obj, opt_crypt_flags,
					hash, 20, out, len);
	}
	if (r < 0) {

		fprintf(stderr, "Compute signature failed: %s\n", sc_strerror(r));
		return 1;
	}

      	for (i=0;i<128;i++) hash_sign[i] = out[i]; 
	hash_sign[128]=0;

	return 0;
}

int get_key(unsigned int usage, sc_pkcs15_object_t **result)
{
	sc_pkcs15_object_t *key, *pin;
	const char	*usage_name;
	sc_pkcs15_id_t	id;
	int		r;

	usage_name = (usage & SC_PKCS15_PRKEY_USAGE_SIGN)? "signature" : "decryption";

	if (certificato != NULL) {
		sc_pkcs15_hex_string_to_id(certificato, &id);
		r = sc_pkcs15_find_prkey_by_id_usage(p15card, &id, usage, &key);

		if (r < 0) {
		// fprintf(stderr, "Unable to find private %s key '%s': %s\n", usage_name, certificato, sc_strerror(r));
		return 2;
		}
	} else {
		r = sc_pkcs15_find_prkey_by_id_usage(p15card, NULL, usage, &key);
		
		if (r < 0) {
		fprintf(stderr, "Unable to find any private %s key: %s\n",
		usage_name, sc_strerror(r));
		return 2;
		}
	}

	*result = key;

	if (key->auth_id.len) {
		static sc_pkcs15_object_t *prev_pin = NULL;
		char	*pincode;

		r = sc_pkcs15_find_pin_by_auth_id(p15card, &key->auth_id, &pin);

		if (r) {
		fprintf(stderr, "Unable to find PIN code for private key: %s\n",
		sc_strerror(r));
		return 1;
		}

		/* Pin already verified previously */
		if (pin == prev_pin)
			return 0;

		pincode = strdup(opt_pincode);
		if (pincode == NULL || *pincode == '\0')
			return 5;

		r = sc_pkcs15_verify_pin(p15card, (struct sc_pkcs15_pin_info *) pin->data,
					(const u8 *) pincode, strlen(pincode));
		
		if (r) {
		fprintf(stderr, "PIN code verification failed: %s\n", sc_strerror(r));
		return 5;
		}

		free(pincode);

		prev_pin = pin;
	}

	return 0;
}


int p15getlistcert()
{       
        int r, i;
        struct sc_pkcs15_object *objs[32];

	r= connect_pcsc();

	if (r)  goto end;

        r = sc_pkcs15_get_objects(p15card, SC_PKCS15_TYPE_CERT_X509, objs, 32);
       
	 if (r < 0) {
                fprintf(stderr, "Certificate enumeration failed: %s\n", sc_strerror(r));
                goto end;
         }

	printf("\n");
	for (i = 0; i < r; i++) {
	unsigned int ii;
        struct sc_pkcs15_cert_info *cert = (struct sc_pkcs15_cert_info *) objs[i]->data;
	
        printf("ID: %s - ", sc_pkcs15_print_id(&cert->id));

        printf("%s - Path: ", objs[i]->label);

        for (ii = 0; ii < cert->path.len; ii++)
                printf("%02X", cert->path.value[ii]);

	printf("\n");

	}

	printf("\n");

end:
 die();

return r;
}                       


int p15getcert(unsigned char *opt_cert)
{       
        int r, i, count;
        struct sc_pkcs15_id id;
        struct sc_pkcs15_object *objs[32];

	id.len = SC_PKCS15_MAX_ID_SIZE;

	r= connect_pcsc();
	if (r)  goto end;

        sc_pkcs15_hex_string_to_id(opt_cert, &id);

        r = sc_pkcs15_get_objects(p15card, SC_PKCS15_TYPE_CERT_X509, objs, 32);
        if (r < 0) {
                fprintf(stderr, "Certificate enumeration failed: %s\n", sc_strerror(r));
                goto end;
        }


        count = r;
        for (i = 0; i < count; i++) {
                struct sc_pkcs15_cert_info *cinfo = (struct sc_pkcs15_cert_info *) objs[i]->data;
                struct sc_pkcs15_cert *cert;
        
                if (sc_pkcs15_compare_id(&id, &cinfo->id) != 1)
                        continue;

                        printf("Reading certificate with ID '%s'\n", opt_cert);

                r = sc_pkcs15_read_certificate(p15card, cinfo, &cert);
                if (r) {
			fprintf(stderr, "Certificate read failed: %s\n", sc_strerror(r));
                        goto end;
                }
		
                r = print_pem_object(cert->data, cert->data_len);

                sc_pkcs15_free_certificate(cert); 
	goto end;

        }

end:
 die();

return r;
}                       

int
print_pem_object(const u8*data, size_t data_len)
{
        unsigned char   *buf = NULL;
        size_t          buf_len = 1024;
        int             r;
	FILE 		*fp_out;
                
        /* With base64, every 3 bytes yield 4 characters, and with
         * 64 chars per line we know almost exactly how large a buffer we
         * will need. */
        buf_len = (data_len + 2) / 3 * 4;
        buf_len += 2 * (buf_len / 64 + 2); /* certain platforms use CRLF */
        buf_len += 64;                     /* slack for checksum etc */
 
        if (!(buf = (unsigned char *) malloc(buf_len))) {
                perror("print_pem_object");
                return 1;
        }
        
        r = sc_base64_encode(data, data_len, buf, buf_len, 64);
        if (r < 0) {
                fprintf(stderr, "Base64 encoding failed: %s\n", sc_strerror(r));
                free(buf);
                return 1;
        }
        
	fp_out = fopen(file_certificato,"w+");
	fprintf(fp_out,"%s","-----BEGIN CERTIFICATE-----\n");
	fprintf(fp_out,"%s",buf);
	fprintf(fp_out,"%s","-----END CERTIFICATE-----\n");

	fclose(fp_out);

        free(buf);

return 0;
}       

int connect_pcsc() {

        int r;
	const char *app_name = "scsign";
	r = sc_establish_context(&ctx, app_name);
        
	if (r) {
             fprintf(stderr, "Failed to establish context: %s\n", sc_strerror(r));
             return r;
        }
         
        r = connect_card(ctx, &card, -1, 0, 0, 0);
        if (r) return r;

        r = sc_pkcs15_bind(card, &p15card);

        if (r) {
                fprintf(stderr, "PKCS#15 initialization failed: %s\n", sc_strerror(r));
                return r;
        }
return 0;
}

void die()
{
        if (card) {
                sc_unlock(card);
                sc_disconnect_card(card, 0);
        }
        if (ctx)
                sc_release_context(ctx);
}

